/*
 * FILE:
 *   enet_network_connection.cpp
 *
 * AUTHOR:
 *   Stephen Thompson, 2008
 *
 * COPYRIGHT:
 *   Public Domain
 *
 */

#include "enet_network_connection.hpp"

#include "../core/coercri_error.hpp"

namespace Coercri {

    EnetNetworkConnection::EnetNetworkConnection(ENetHost *host, const std::string &hostname, int port)
        : peer(0), state(PENDING)
    {
        ENetAddress address;
        enet_address_set_host(&address, hostname.c_str());
        address.port = port;

        peer = enet_host_connect(host, &address, 1);  // open 1 channel only
        if (!peer) {
            throw CoercriError("EnetNetworkConnection: enet_host_connect failed");
        }

        peer->data = this;
    }

    EnetNetworkConnection::EnetNetworkConnection(ENetPeer *peer_)
        : peer(peer_), state(CONNECTED)
    {
        peer->data = this;
    }

    EnetNetworkConnection::~EnetNetworkConnection()
    {
        while (!queued_packets.empty()) {
            enet_packet_destroy(queued_packets.front());
            queued_packets.pop();
        }
        if (peer) {
            // Use enet_peer_disconnect_now rather than enet_peer_reset
            // (the only real difference is that disconnect_now will take
            // one last shot at sending a DISCONNECT message before
            // resetting).
            enet_peer_disconnect_now(peer, 0);
            peer->data = 0;
        }
    }

    EnetNetworkConnection::State EnetNetworkConnection::getState() const
    {
        return state;
    }

    void EnetNetworkConnection::close()
    {
        if (peer && state != CLOSED) {
            state = CLOSED;
            enet_peer_disconnect(peer, 0);
        }
        // NOTE: Keep the link to the peer open (ie dont zero out peer->data and peer),
        // as the peer is still "active" while we wait for the disconnect message.
    }

    void EnetNetworkConnection::transmit(const std::vector<ubyte> &buf)
    {
        if (state == CONNECTED && peer) {
            ENetPacket *packet = enet_packet_create(&buf[0], buf.size(), ENET_PACKET_FLAG_RELIABLE);
            enet_peer_send(peer, 0, packet);
        }
    }

    void EnetNetworkConnection::receive(std::vector<ubyte> &buf)
    {
        if (!queued_packets.empty()) {
            ubyte* ptr = static_cast<ubyte*>(queued_packets.front()->data);
            buf.assign(ptr, ptr + queued_packets.front()->dataLength);
            enet_packet_destroy(queued_packets.front());
            queued_packets.pop();
        }
    }

    void EnetNetworkConnection::onReceiveAcknowledgment()
    {
        state = CONNECTED;
    }

    void EnetNetworkConnection::onReceivePacket(ENetPacket *packet)
    {
        queued_packets.push(packet);
    }

    void EnetNetworkConnection::onDisconnect()
    {
        state = CLOSED;

        // Break the link to the peer.
        peer->data = 0;
        peer = 0;
    }
}
