/*
 * FILE:
 *   window_listener.hpp
 *
 * PURPOSE:
 *   Interface for window events: close, gain/lose focus,
 *   keyboard/mouse input, etc.
 *
 * AUTHOR:
 *   Stephen Thompson, 2008.
 *
 * COPYRIGHT:
 *   Public Domain.
 *
 */

#ifndef COERCRI_WINDOW_LISTENER_HPP
#define COERCRI_WINDOW_LISTENER_HPP

#include "cooked_key.hpp"
#include "mouse_button.hpp"
#include "raw_key.hpp"

namespace Coercri {

    class GfxContext;
    class Region;
    class Window;
    
    class WindowListener {
    public:

        virtual ~WindowListener() { }
        
        // Repaint required.
        // The GfxContext's clip rectangle will be set to the entire
        // window, note however that only the given region will
        // actually need to be repainted.
        virtual void onRepaint(GfxContext &gfx_context, const Region &region) { }

        // Window closed
        virtual void onClose() { }

        // Window gained or lost focus
        virtual void onGainFocus() { }
        virtual void onLoseFocus() { }

        // Window resized
        virtual void onResize(int new_width, int new_height) { }

        // Raw key events
        virtual void onRawKeyDown(RawKey key) { }
        virtual void onRawKeyUp(RawKey key) { }

        // Cooked key events.
        // For normal text: key==CK_CHARACTER, character is a Unicode code point.
        // For special keys: key!=CK_CHARACTER, character==0.
        // In all cases: modifiers is set to an OR'd combination of Modifier flags (see cooked_key.hpp).
        virtual void onCookedKey(CookedKey key, int character, int modifiers) { }
        
        // Mouse button events
        virtual void onMouseDown(int x, int y, MouseButton button) { }
        virtual void onMouseUp(int x, int y, MouseButton button) { }

        // Mouse motion event
        virtual void onMouseMove(int new_x, int new_y) { }
    };

}

#endif
