/*
 * gfx_manager.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "gfx_manager.hpp"
#include "graphic_transform.hpp"
#include "load_font.hpp"
#include "my_exceptions.hpp"
#include "rstream/rstream.hpp"

#include "coercri/gfx/load_bmp.hpp"

GfxManager::GfxManager(shared_ptr<Coercri::GfxDriver> gfx_driver_,
                       shared_ptr<Coercri::TTFLoader> ttf_loader_,
                       const vector<string> &ttf_font_names_,
                       const vector<string> &bmp_font_names_)
    : gfx_driver(gfx_driver_),
      ttf_loader(ttf_loader_),
      ttf_font_names(ttf_font_names_),
      bmp_font_names(bmp_font_names_),
      font_size(0),
      npix(0)
{
}

GfxManager::~GfxManager()
{
    deleteAllGraphics();
}

shared_ptr<Coercri::Font> GfxManager::getFont()
{
    return font;
}

void GfxManager::setFontSize(int new_size)
{
    if (font_size == new_size) return;
    font = LoadFont(*ttf_loader, ttf_font_names, bmp_font_names, new_size);
    font_size = new_size;
}  

void GfxManager::serializeGraphic(Coercri::NetworkConnection &remote, const Coercri::Graphic &g, int id)
{
    const GfxRevMap::const_iterator it = gfx_rev_map.find(&g);
    if (it == gfx_rev_map.end()) throw UnexpectedError("graphic serialization failed");

    remote.putString(it->second.filename);
    remote.putCInt(it->second.hx);
    remote.putCInt(it->second.hy);
    remote.putCInt(it->second.r);
    remote.putCInt(it->second.g);
    remote.putCInt(it->second.b);

    pair<IDMap::iterator, bool> p = id_map.insert(make_pair(&g, id));
    if (!p.second) throw UnexpectedError("graphic serialized twice");
}

Coercri::Graphic * GfxManager::createGraphicFromNetwork(Coercri::NetworkConnection &remote)
{
    string filename = remote.getString();
    int hx = remote.getCInt();
    int hy = remote.getCInt();
    int r = remote.getCInt();
    int g = remote.getCInt();
    int b = remote.getCInt();
    return createGraphicFromFile(filename, hx, hy, r, g, b);
}

int GfxManager::getGraphicID(const Coercri::Graphic &g) const
{
    const IDMap::const_iterator it = id_map.find(&g);
    if (it == id_map.end()) {
        // This only happens if they try to get ID of a graphic that doesnt have an ID.
        // Callers shouldn't be doing this, so assert in debug build, but try to carry on and return 0 in release mode.
        ASSERT(false);
        return 0;
    } else {
        return it->second;
    }
}

Coercri::Graphic * GfxManager::createGraphicFromFile(const string &filename, int hx, int hy, int r, int g, int b)
{
    // Note: We don't (yet) bother to optimize the case where the same filename,
    // but a different hx/hy/r/g/b is being loaded. The BMP file is loaded twice in that case.
    GraphicSpec gs = { filename, hx, hy, r, g, b };
    GfxMap::const_iterator it = gfx_map.find(gs);
    if (it != gfx_map.end()) {
        return it->second.get();
    } else {
        // Open the file. Only BMP format supported for now.
        RStream str(filename);
        shared_ptr<const Coercri::PixelArray> pixels = Coercri::LoadBMP(str);

        // Apply colour key if necessary
        if (r != -1) {
            ColourChange cc;
            cc.add(Coercri::Color(r, g, b, 255), Coercri::Color(r, g, b, 0));
            pixels = CreateGraphicWithCC(pixels, cc);
        }
        
        // Create the graphic
        shared_ptr<Coercri::Graphic> new_graphic
            = gfx_driver->createGraphic(pixels, hx, hy);
        
        // Add to the map and return
        gfx_map.insert(GfxMap::value_type(gs, new_graphic));
        gfx_rev_map.insert(GfxRevMap::value_type(new_graphic.get(), gs));
        return new_graphic.get();
    }
}

void GfxManager::deleteAllGraphics()
{
    // Clearing gfx_map will release all the shared_ptr references to
    // the graphics and therefore will delete them all
    gfx_map.clear();
    gfx_rev_map.clear();
    id_map.clear();

    // We also need to delete any cached transformed-graphics
    cached_gfx_map.clear();
    recent_list.clear();
    npix = 0;
}

void GfxManager::setGfxResizer(shared_ptr<GfxResizer> resizer)
{
    gfx_resizer = resizer;
    // We need to delete any cached graphics, as they may not be valid with the new resizer
    cached_gfx_map.clear();
    recent_list.clear();
    npix = 0;
}

shared_ptr<GfxResizer> GfxManager::getGfxResizer()
{
    return gfx_resizer;
}

shared_ptr<Coercri::Graphic> GfxManager::createGraphic(const GraphicKey &key)
{
    // NOTE: This routine should be called with either a resize or a colour-change,
    // but not both at the same time.

    shared_ptr<const Coercri::PixelArray> new_pixels;
    int new_hx, new_hy;
    key.original->getHandle(new_hx, new_hy);
    
    if (key.cc.empty()) {
        // Resize only
        ASSERT(key.new_width != key.original->getWidth() || key.new_height != key.original->getHeight());

        // Call the resizing routine
        new_pixels = CreateResizedGraphic(*gfx_resizer, key.original->getPixels(), key.new_width, key.new_height,
                                          new_hx, new_hy, new_hx, new_hy);

    } else {
        // Colour change only
        ASSERT(key.new_width == key.original->getWidth() && key.new_height == key.original->getHeight());

        // Call the colour-changing routine
        new_pixels = CreateGraphicWithCC(key.original->getPixels(), key.cc);
    }
    
    shared_ptr<Coercri::Graphic> new_graphic(gfx_driver->createGraphic(new_pixels, new_hx, new_hy));
    return new_graphic;
}

const Coercri::Graphic & GfxManager::getGraphic(const GraphicKey &key)
{
    CachedGfxMap::iterator it = cached_gfx_map.find(key);

    if (it != cached_gfx_map.end()) {
        // It's in the cache. Bump it to the front of the MRU list
        recent_list.splice(recent_list.begin(), recent_list, it->second.list_position);
        return *it->second.new_graphic;

    } else {
        // Not in cache. Need to create a new graphic
        shared_ptr<Coercri::Graphic> new_graphic = createGraphic(key);

        // Add to the RecentList (at the front)
        recent_list.push_front(key);

        // Add to the main cache
        GraphicData gd;
        gd.new_graphic = new_graphic;
        gd.list_position = recent_list.begin();
        cached_gfx_map.insert(make_pair(key, gd));

        // Increase npix
        npix += key.new_width * key.new_height;

        // Delete old graphics if necessary
        deleteOld();

        // Return the new graphic
        return *new_graphic;
    }
}

bool GfxManager::tooBig() const
{
    // NOTE: minimum_number should be at least 2, otherwise getResizedGraphicWithCC
    // may fail (because its intermediate "gfx_cc" may get deleted before we can use it!).
    
    const int minimum_number = 5;
    const int maximum_number = 1000;
    const int maximum_pixels = 1000000;

    const int current_number = cached_gfx_map.size();

    if (current_number <= minimum_number) return false;
    if (current_number > maximum_number) return true;
    if (npix > maximum_pixels) return true;
    return false;
}

void GfxManager::deleteOld()
{
    // Delete old graphic(s) if the cache is too big.
    while (tooBig()) {
        ASSERT(!recent_list.empty());

        // Find the least recently accessed graphic in the CachedGfxMap. We are going to remove it.
        CachedGfxMap::iterator it = cached_gfx_map.find(recent_list.back());
        ASSERT(it != cached_gfx_map.end());
        
        // Reduce npix accordingly
        npix -= it->first.new_width * it->first.new_height;

        // Remove it from CachedGfxMap and RecentList
        // (This will also delete the graphic, since the shared_ptr will be released)
        cached_gfx_map.erase(it);
        recent_list.pop_back();
    }
}

const Coercri::Graphic & GfxManager::getGraphicWithCC(const Coercri::Graphic &original, const ColourChange &cc)
{
    // Short cut:
    if (cc.empty()) return original;

    // Look it up in cache (creating if necessary):
    GraphicKey key;
    key.original = &original;
    key.cc = cc;
    key.new_width = original.getWidth();
    key.new_height = original.getHeight();
    return getGraphic(key);
}

const Coercri::Graphic & GfxManager::getResizedGraphic(const Coercri::Graphic &original, int new_width, int new_height)
{
    // Short cut:
    if (new_width == original.getWidth() && new_height == original.getHeight()) return original;

    // Look it up in cache (creating if necessary):
    GraphicKey key;
    key.original = &original;
    key.cc = ColourChange();
    key.new_width = new_width;
    key.new_height = new_height;
    return getGraphic(key);
}

void GfxManager::drawTransformedGraphic(Coercri::GfxContext &gc, int x, int y,
                                        const Coercri::Graphic &gfx, int new_width, int new_height)
{
    const Coercri::Graphic & gfx_resized = getResizedGraphic(gfx, new_width, new_height);
    gc.drawGraphic(x, y, gfx_resized);
}

void GfxManager::drawTransformedGraphic(Coercri::GfxContext &gc, int x, int y,
                                        const Coercri::Graphic &gfx, int new_width, int new_height,
                                        const ColourChange &cc)
{
    const Coercri::Graphic & gfx_cc = getGraphicWithCC(gfx, cc);
    drawTransformedGraphic(gc, x, y, gfx_cc, new_width, new_height);
}
