/*
 * loading_screen.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "client_config.hpp"
#include "knights_game.hpp"
#include "loading_screen.hpp"
#include "menu_screen.hpp"
#include "rstream/rstream.hpp"
#include "server_config.hpp"
#include "sound_manager.hpp"
#include "x_centre.hpp"

#include "coercri/gfx/window_listener.hpp"

namespace {
    class LoadingListener : public Coercri::WindowListener {
    public:
        explicit LoadingListener(shared_ptr<Coercri::Font> f) : font(f) { }
        virtual void onRepaint(Coercri::GfxContext &gc, const Coercri::Region &rgn);
    private:
        shared_ptr<Coercri::Font> font;
    };

    void LoadingListener::onRepaint(Coercri::GfxContext &gc, const Coercri::Region &rgn)
    {
        gc.clearScreen(Coercri::Color(0,0,0));
        XCentre(gc, *font, font->getTextHeight() + 15, "LOADING");
    }

    void SoundCallback(const string &name, void *cxt)
    {
        SoundManager *sm = static_cast<SoundManager*>(cxt);
        sm->loadSound(name);
    }
    
    shared_ptr<ServerConfig> LoadServerConfig(GfxManager &gm, SoundManager &sm)
    {
        shared_ptr<ServerConfig> server_config(new ServerConfig(gm));
        server_config->loadConfigFile("main.txt", &SoundCallback, &sm);
        return server_config;
    }
    
    void SendServerConfig(Coercri::NetworkConnection &remote, shared_ptr<ServerConfig> server_config)
    {
        server_config->serializeAll(remote);
        remote.endPacket();
    }

    shared_ptr<ClientConfig> LoadClientConfig(Coercri::NetworkConnection &remote, GfxManager &gm)
    {
        if (!remote.inAvail()) {
            return shared_ptr<ClientConfig>();
        } else {
            shared_ptr<ClientConfig> client_config(new ClientConfig(remote, gm));
            return client_config;
        }
    }
}

bool LoadingScreen::start(KnightsGame &kg, shared_ptr<Coercri::Window> win, gcn::Gui &gui)
{
    knights_game = &kg;
    window = win;
    window_listener.reset(new LoadingListener(kg.getFont()));
    window->addWindowListener(window_listener.get());
    return false;
}

LoadingScreen::~LoadingScreen()
{
    window->rmWindowListener(window_listener.get());
}

void LoadingScreen::update()
{
    if (knights_game->server) {
        knights_game->server_config = LoadServerConfig(knights_game->getGfxManager(), knights_game->getSoundManager());
        if (knights_game->remote) {
            SendServerConfig(*knights_game->remote, knights_game->server_config);
        }
    } else {
        ASSERT(knights_game->remote);
        knights_game->client_config = LoadClientConfig(*knights_game->remote, knights_game->getGfxManager());
        if (!knights_game->client_config) {
            // The client config has not yet been loaded. We should wait a bit and then
            // retry (at next update point).
            return;
        }
    }

    // If we get to here then we are ready to proceed to the MenuScreen....
    auto_ptr<Screen> menu_screen(new MenuScreen);
    knights_game->requestScreenChange(menu_screen);
}
