/*
 * local_dungeon_view.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "control.hpp"
#include "controller.hpp"
#include "draw.hpp"
#include "gfx_manager.hpp"
#include "local_dungeon_view.hpp"
#include "mini_map_colour.hpp"
#include "round.hpp"
#include "sound_manager.hpp"

#include "coercri/gfx/rectangle.hpp"

// NB item_depth must be greater than the other depths here.
const int icon_depth = -205;
const int entity_depth = -200;   // note this is modified (in steps of 10) for MapHeight, and 1 is subtracted for weapon overlays.
const int item_depth = -100;

const int MY_X_OFFSCREEN_VALUE = -999;

namespace {
    void Resize(int ref_left, int ref_top, int ref_blk_sz, int ref_x_blks, int ref_y_blks,
                float scale, int left, int top,
                int &phy_left, int &phy_top, int &phy_blk_sz)
    {
        phy_left = Round(float(ref_left) * scale) + left;
        phy_top = Round(float(ref_top) * scale) + top;
        const float phy_blk_sz_f = ref_blk_sz * scale;
        phy_blk_sz = int(phy_blk_sz_f);  // round down
        const float leftover = phy_blk_sz_f - phy_blk_sz;
        phy_left += int(leftover * ref_x_blks) / 2;
        phy_top += int(leftover * ref_y_blks) / 2;
    }
}

LocalDungeonView::LocalDungeonView(const Controller & ctrlr, const Controller *ctrlr2, SoundManager &sm)
    : gvt(0), current_room(-1),
      my_x(MY_X_OFFSCREEN_VALUE), my_y(0), my_facing(D_NORTH),
      my_approached(false), 
      menu_gfx_centre(0), menu_gfx_empty(0),
      controller(ctrlr), controller2(ctrlr2),
      tap_control(0), menu_open(false), menu_highlight(false),
      fire_start_time(0), attack_mode(false), menu_null(M_OK),
      last_known_time(0), last_known_x(0), last_known_y(0),
      last_mtime(0), flash_screen_start(-9999),
      skull_renderer(0), potion_renderer(0),
      nskulls(0), health(0), magic(NO_POTION), poison_immun(false),
      won(false), lost(false),
      cheat_controls(0),
      sound_manager(sm)
{
    for (int i=0; i<4; ++i) menu_control[i] = 0;
}

//
// modifiers
//

void LocalDungeonView::setMiniMap(int w, int h)
{
    mini_map.create(w, h);
}

void LocalDungeonView::setCurrentRoom(int r, int w, int h)
{
    typedef map<int, RoomData>::iterator Iter;

    // Wipe all entities from the current map
    RoomData &rd(rooms[current_room]);
    rd.emap.clear();

    // Create the new map (if necessary) and reset current_room & current_dmap.
    Iter it = rooms.find(r);
    if (it == rooms.end()) {
        pair<Iter,bool> p = rooms.insert(make_pair(r, RoomData()));
        ASSERT(p.second);
        it = p.first;
        it->second.create(w, h);
    }
    current_room = r;
}

int LocalDungeonView::getCurrentRoom() const
{
    return current_room;
}

void LocalDungeonView::updateKnightIndicator()
{
    // Set knight indicator (highlight 0)

    map<int,RoomData>::iterator ri = rooms.find(current_room);
    if (my_x == MY_X_OFFSCREEN_VALUE || ri == rooms.end()
    || ri->second.origin.isNull()) {
        // Knight not found, OR current room is not mapped
        // Don't draw a highlight.
        mini_map.setHighlight(MapCoord(), 0);
    } else {
        // Knight exists and current room is mapped
        // Draw the highlight.
        mini_map.setHighlight(MapCoord(ri->second.origin.getX()+my_x,
                                       ri->second.origin.getY()+my_y), 0);
    }
}

void LocalDungeonView::mapCurrentRoom(const MapCoord &tl_pos)
{
    // Set the "origin" field in the current RoomData.
    map<int,RoomData>::iterator it = rooms.find(current_room);
    if (it == rooms.end()) return;
    it->second.origin = tl_pos;

    // Copy it to the mini-map if necessary.
    if (it->second.awaiting_copy) {
        copyRoomToMiniMap(it->second);
    }
    
    // Set the knight indicator (highlight 0).
    updateKnightIndicator();
}

void LocalDungeonView::mapWall(const MapCoord &mc)
{
    mini_map.setColour(mc, COL_WALL);
}

void LocalDungeonView::mapKnightLocation(int id, const MapCoord &pos)
{
    // We add 1 to the ID, in order to prevent clashes with ID 0 which is always the
    // current knight's location.
    mini_map.setHighlight(pos, id+1);
}

void LocalDungeonView::mapItemLocation(const MapCoord &mc, bool on)
{
    if (mc.isNull()) return;
    const int id = mc.getY() * mini_map.getHeight() + mc.getX() + 10; // add 10 to prevent clashes with other IDs
    if (on) {
        mini_map.setHighlight(mc, id);
    } else {
        mini_map.setHighlight(MapCoord(), id);
    }
}

void LocalDungeonView::wipeMap()
{
    // Wipe memorized room positions
    for (map<int, RoomData>::iterator it = rooms.begin(); it != rooms.end(); ++it) {
        it->second.origin = MapCoord();
        it->second.awaiting_copy = true;
    }

    // Wipe the mini-map itself
    mini_map.wipe();
}

bool LocalDungeonView::isSquareMapped(const MapCoord &mc) const
{
    if (mini_map.getColour(mc) == COL_UNMAPPED) return false; else return true;
}

void LocalDungeonView::copyRoomToMiniMap(RoomData &rd)
{
    if (!rd.awaiting_copy) return;
    if (rd.origin.isNull()) return;
    
    rd.awaiting_copy = false;
    const int ox = rd.origin.getX();
    const int oy = rd.origin.getY();

    for (int i=0; i<rd.width; ++i) {
        for (int j=0; j<rd.height; ++j) {
            const list<RoomData::GfxEntry> & gfx(rd.lookupGfx(i,j));
            MiniMapColour col = COL_FLOOR;
            for (list<RoomData::GfxEntry>::const_iterator it = gfx.begin(); it != gfx.end();
            ++it) {
                if (it->col != COL_UNMAPPED) {
                    col = it->col;
                    break;
                }
            }
            mini_map.setColour(MapCoord(i+ox, j+oy), col);
        }
    }
}

void LocalDungeonView::addEntity(unsigned short int id, int x, int y, MapHeight ht, MapDirection facing,
                                 const Anim * anm, const Overlay * ovr, int af, int atz_diff,
                                 bool ainvuln,
                                 int cur_ofs, MotionType mt, int mtr)
{
    map<int,RoomData>::iterator ri = rooms.find(current_room);
    if (ri == rooms.end()) return;
    ri->second.emap.addEntity(gvt, id, x, y, ht, facing, anm, ovr, af, atz_diff+gvt, ainvuln,
                              cur_ofs, mt, mtr);
    if (id==0) {
        my_x = x;
        my_y = y;
        if (my_x >= 0) {
            last_known_x = my_x;
            last_known_y = my_y;
        }
        my_approached = false;
        updateKnightIndicator();
    }
}

void LocalDungeonView::rmEntity(unsigned short int id)
{
    map<int,RoomData>::iterator ri = rooms.find(current_room);
    if (ri == rooms.end()) return;
    ri->second.emap.rmEntity(id);
    if (id==0) {
        my_x = MY_X_OFFSCREEN_VALUE;
        updateKnightIndicator();
    }
}

void LocalDungeonView::repositionEntity(unsigned short int id, int new_x, int new_y)
{
    map<int,RoomData>::iterator ri = rooms.find(current_room);
    if (ri == rooms.end()) return;
    ri->second.emap.repositionEntity(id, new_x, new_y);
    if (id == 0) {
        my_x = new_x;
        my_y = new_y;
        if (my_x >= 0) {
            last_known_x = my_x;
            last_known_y = my_y;
        }
        my_approached = false;
        updateKnightIndicator();
    }
}

void LocalDungeonView::moveEntity(unsigned short int id, MotionType mt, int motion_dur, bool missile_mode)
{
    map<int,RoomData>::iterator ri = rooms.find(current_room);
    if (ri == rooms.end()) return;
    ri->second.emap.moveEntity(gvt, id, mt, motion_dur, missile_mode);
    if (id==0) {
        my_approached = (mt == MT_APPROACH);
        updateKnightIndicator();
    }
}

void LocalDungeonView::flipEntityMotion(unsigned short int id, int initial_delay, int motion_dur)
{
    map<int,RoomData>::iterator ri = rooms.find(current_room);
    if (ri == rooms.end()) return;
    ri->second.emap.flipEntityMotion(gvt, id, initial_delay, motion_dur);
    if (id == 0) {
        // not sure if updateKnightIndicator is strictly necessary here,
        // but we'll have it anyway
        updateKnightIndicator();
    }
}   

void LocalDungeonView::setAnimData(unsigned short int id, const Anim *a, const Overlay *o,
                                   int af, int atz_diff, bool ainvuln, bool currently_moving)
{
    map<int,RoomData>::iterator ri = rooms.find(current_room);
    if (ri == rooms.end()) return;
    ri->second.emap.setAnimData(id, a, o, af, atz_diff ? atz_diff+gvt : 0, ainvuln, currently_moving);
}

void LocalDungeonView::setFacing(unsigned short int id, MapDirection new_facing)
{
    map<int,RoomData>::iterator ri = rooms.find(current_room);
    if (ri == rooms.end()) return;  
    ri->second.emap.setFacing(id, new_facing);
    if (id==0) {
        updateKnightIndicator();
    }
}

void LocalDungeonView::clearTiles(int x, int y)
{
    map<int,RoomData>::iterator ri = rooms.find(current_room);
    if (ri == rooms.end()) return;
    if (!ri->second.valid(x,y)) return;
    list<RoomData::GfxEntry> &gfx(ri->second.lookupGfx(x, y));
    // We know that the tiles come before items or other stuff like that.
    while (!gfx.empty() && gfx.front().depth != item_depth) gfx.pop_front();
}

void LocalDungeonView::setTile(int x, int y, int depth, const Coercri::Graphic * gfx,
                               shared_ptr<const ColourChange> cc, MiniMapColour col)
{
    map<int,RoomData>::iterator ri = rooms.find(current_room);
    if (ri == rooms.end()) return;
    if (!ri->second.valid(x,y)) return;
    list<RoomData::GfxEntry> &glst(ri->second.lookupGfx(x, y));
    list<RoomData::GfxEntry>::iterator it = glst.begin();
    while (1) {
        // The list is in sorted order, highest depth first.
        if (it == glst.end() || depth > it->depth) {
            if (gfx) {
                RoomData::GfxEntry ge;
                ge.gfx = gfx;
                ge.cc = cc;
                ge.col = col;
                ge.depth = depth;
                glst.insert(it, ge);
            }
            break;
        } else if (it->depth == depth) {
            if (gfx) {
                it->gfx = gfx;
                it->cc = cc;
                it->col = col;
            } else {
                glst.erase(it);
            }
            break;
        } else {
            ++it;
        }
    }
    
    ri->second.awaiting_copy = true;
}

void LocalDungeonView::setItem(int x, int y, const Coercri::Graphic * graphic)
{
    // We can implement this using setTile.
    setTile(x, y, item_depth, graphic, shared_ptr<const ColourChange>(), COL_UNMAPPED);
}


void LocalDungeonView::placeIcon(int x, int y, const Coercri::Graphic *g, int dur)
{
    if (current_room == -1) return;
    
    // Add the 'icon' to the gmap. Also store in 'icons' list.
    setTile(x, y, icon_depth, g, shared_ptr<const ColourChange>(), COL_UNMAPPED);
    LocalIcon ic;
    ic.room_no = current_room;
    ic.x = x;
    ic.y = y;
    ic.expiry = dur + gvt;
    icons.push(ic);
}

void LocalDungeonView::setBackpack(int slot, const Coercri::Graphic *gfx, const Coercri::Graphic *overdraw, int no_carried, int no_max)
{
    BackpackEntry be;
    be.gfx = gfx;
    be.overdraw = overdraw;
    be.no_carried = no_carried;
    be.no_max = no_max;
    backpack.erase(slot);
    backpack.insert(make_pair(slot, be));
}

void LocalDungeonView::addSkull()
{
    ++nskulls;
}

void LocalDungeonView::setHealth(int h)
{
    health = h;
}

void LocalDungeonView::setPotionMagic(PotionMagic pm, bool pi)
{
    magic = pm;
    poison_immun = pi;
}

void LocalDungeonView::flashMessage(const string &s, int ntimes)
{
    Message m;
    m.message = s;
    m.start_time = gvt;
    m.stop_time = gvt + (CfgInt("message_on_time") + CfgInt("message_off_time"))*ntimes;
    messages.push_back(m);
}

void LocalDungeonView::cancelContinuousMessages()
{
    cts_messages.clear();
}

void LocalDungeonView::addContinuousMessage(const string &s)
{
    cts_messages.push_back(s);
}

void LocalDungeonView::flashScreen(int delay)
{
    flash_screen_start = gvt + delay;
}

void LocalDungeonView::playSound(const string &name, int frequency, bool &flag)
{
    // flag will be true if this sound has already been played on this machine
    if (!flag) {
        sound_manager.playSound(name, frequency);
    }
    flag = true;  // set the flag.
}


//
// Menus/Controls
//

void LocalDungeonView::setMenuControl(MapDirection d, const Control *ctrl,
                                      const Control *prev)
{
    if (menu_null==M_CTS && menu_null_dir==d && ctrl != prev) {
        // This stops e.g. immediately opening a door after the lock picks completed.
        menu_null = M_NULL;
    }
    menu_control[d] = ctrl;
}

void LocalDungeonView::setAvailableControls(const vector<pair<const Control *,bool> >
                                            &controls)
{
    const Control * prev_menu_control[4];
    for (int i=0; i<4; ++i) {
        prev_menu_control[i] = menu_control[i];
        menu_control[i] = 0;
    }
    tap_control = 0;
    int tap_pri = 0;

    // the server is currently telling us that all controls in "controls" are currently
    // possible and available to our knight. so put those controls into the menu or
    // "tap_control" as appropriate.

    for (vector<pair<const Control *, bool> >::const_iterator ctrl = controls.begin(); 
    ctrl != controls.end(); ++ctrl) {
        // Find the (primary) control with the highest tap priority:
        // (NB: control is primary iff ctrl->second == true.)
        if (ctrl->first->getTapPriority() > tap_pri && ctrl->second) {
            tap_pri = ctrl->first->getTapPriority();
            tap_control = ctrl->first;
        }
        if (ctrl->first->getMenuGraphic() != 0) {
            MapDirection d = ctrl->first->getMenuDirection();
            // If no control in direction d, or if the control in direction d has
            // property MS_WEAK, then save ctrl->first into menu position d.
            if (menu_control[d] == 0 ||
            (menu_control[d]->getMenuSpecial() & Control::MS_WEAK)) {
                setMenuControl(d, ctrl->first, prev_menu_control[d]);
            }
        }
    }

    // Now try to place controls that could not be put in their preferred position.
    for (vector<pair<const Control *, bool> >::const_iterator ctrl = controls.begin(); 
    ctrl != controls.end(); ++ctrl) {
        if (ctrl->first->getMenuGraphic() != 0) {
            MapDirection d = ctrl->first->getMenuDirection();
            if (menu_control[d] != ctrl->first &&
            (ctrl->first->getMenuSpecial() & Control::MS_WEAK) == 0) {
                for (int i=0; i<4; ++i) {
                    if (menu_control[i] == 0) {
                        setMenuControl(MapDirection(i), ctrl->first, prev_menu_control[i]);
                        break;
                    }
                }
            }
        }
    }
}

void LocalDungeonView::setMenuHighlight(const Control *ctrl)
{
    if (ctrl) {
        for (int i=0; i<4; ++i) {
            if (menu_control[i] == ctrl) {
                menu_highlight = true;
                menu_highlight_dir = MapDirection(i);
                return;
            }
        }
    }
    menu_highlight = false;
}

const Control * LocalDungeonView::readController()
{
    // This routine reads the joystick/controller, and converts this
    // information to an appropriate "Control". It also displays the
    // on-screen menu (when fire is held down) if necessary.

    // set up some constants
    const int menu_delay = CfgInt("menu_delay"); // cutoff btwn 'tapping' and 'holding' fire.
    const bool approached = my_approached;

    // get current controller state
    MapDirection ctrlr_dir;
    bool ctrlr_centred, ctrlr_fire, suicide_keys;
    controller.get(ctrlr_dir, ctrlr_centred, ctrlr_fire, suicide_keys);

    if (controller2) {
        MapDirection ctrlr_dir_2;
        bool ctrlr_centred_2 = true, ctrlr_fire_2 = false, suicide_keys_2 = false;
        controller2->get(ctrlr_dir_2, ctrlr_centred_2, ctrlr_fire_2, suicide_keys_2);
        if (ctrlr_centred && !ctrlr_centred_2) {
            ctrlr_dir = ctrlr_dir_2;
            ctrlr_centred = false;
        }
        ctrlr_fire = ctrlr_fire || ctrlr_fire_2;
        suicide_keys = suicide_keys || suicide_keys_2;
    }
    
    // if suicide keys held, then we can exit immediately...
    if (suicide_keys) return g_standard_controls[SC_SUICIDE];

    // check the cheat keys:
    int k = controller.getCheatKey();
    if (cheat_controls && k >= 1 && k <= cheat_controls->size()) {
        return (*cheat_controls)[k-1];
    }
    
    // work out 'long' fire state
    enum FireState { NONE, TAPPED, HELD } long_fire = NONE;
    if (ctrlr_fire) {
        if (fire_start_time == 0) {
            // fire_start_time: set to time when fire was pressed down (or 0 if fire is
            // currently released)
            // attack_mode: if fire+direction is used to attack, then attack_mode is set true,
            // which disables menus and 'tapping'.
            fire_start_time = gvt;
            attack_mode = false;
        }
        if (gvt >= fire_start_time + menu_delay && !attack_mode) {
            long_fire = HELD;
        }
    } else {
        // do a 'tap' if allowed
        if (fire_start_time > 0 && fire_start_time < gvt + menu_delay && !attack_mode) {
            long_fire = TAPPED;
        }
        // reset fire state when fire is released
        fire_start_time = 0;
        attack_mode = false;
    }

    
    if (menu_open) {
        // Menu open

        if (long_fire != HELD) {
            // close the menu
            menu_open = false;
            menu_null = M_OK;
        } else {
            if (menuNull() && (ctrlr_centred || ctrlr_dir != menu_null_dir)) {
                // Cancel null dir.
                menu_null = M_OK;
            }
            if (!ctrlr_centred && (!menuNull() || ctrlr_dir != menu_null_dir)) {
                // A menu option has been selected
                const Control *c = menu_control[ctrlr_dir];
                if (c) {
                    // stop the action being executed more than once in a row.
                    // (continuous actions will be stopped only if the control disappears,
                    // which is what M_CTS is for.)
                    menu_null = (c->isContinuous()? M_CTS: M_NULL);
                    menu_null_dir = ctrlr_dir;
                }
                return c;
            }
        }

    } else if (approached) {
        // Approached, and menu closed
        if ((ctrlr_centred || ctrlr_dir != my_facing) && !ctrlr_fire) {
            // Withdraw if controller is no longer pointing in the approach direction
            // and fire is released.
            return g_standard_controls[SC_WITHDRAW];
        } else {
            if (long_fire == HELD) {
                menu_open = true;
                menu_null = M_NULL;
                menu_null_dir = my_facing;
            } else if (long_fire == TAPPED) { 
                return tap_control;
            }
        }
    } else {
        // Not approached, and menu closed
        if (ctrlr_centred) {
            if (long_fire == HELD) {
                menu_open = true;
                menu_null = M_OK;
            } else if (long_fire == TAPPED) { 
                return tap_control;
            }
        } else {
            if (ctrlr_fire) {
                attack_mode = true;
                return g_standard_controls[SC_ATTACK + ctrlr_dir];
            } else {
                my_facing = ctrlr_dir;
                return g_standard_controls[SC_MOVE + ctrlr_dir];
            }
        }
    }

    // No commands coming from the controller at present.
    return 0;
}



//
// drawing
//

void LocalDungeonView::draw(Coercri::GfxContext &gc, int left, int top,
                            GfxManager &gm, float scale, int new_gvt)
{
    gvt = new_gvt;

    static const int num_inventory_slots = CfgInt("dpy_inventory_slots");
    static const int ref_inventory_left = CfgInt("dpy_inventory_left");
    static const int ref_inventory_top = CfgInt("dpy_inventory_top");
    static const int ref_inventory_width = CfgInt("dpy_inventory_width");
    static const int ref_inventory_height = CfgInt("dpy_inventory_height");
    static const int ref_inventory_spacing = CfgInt("dpy_inventory_spacing");
    static const int ref_inventory_gem_height = CfgInt("dpy_inventory_gem_height");
    static const int ref_mini_map_left = CfgInt("dpy_map_left");
    static const int ref_mini_map_top = CfgInt("dpy_map_top");
    static const int ref_mini_map_width = CfgInt("dpy_map_width");
    static const int ref_mini_map_height = CfgInt("dpy_map_height");
    static const int ref_skulls_left = CfgInt("dpy_skulls_left");
    static const int ref_skulls_top = CfgInt("dpy_skulls_top");
    static const int ref_potion_left = CfgInt("dpy_potion_left");
    static const int ref_potion_top = CfgInt("dpy_potion_top");

    static const int ref_pixels_per_square = CfgInt("dpy_pixels_per_square");
    static const int ref_dungeon_left = CfgInt("dpy_dungeon_left");
    static const int ref_dungeon_top = CfgInt("dpy_dungeon_top");
    static const int dungeon_tiles_x = CfgInt("dpy_dungeon_width") / ref_pixels_per_square;
    static const int dungeon_tiles_y = CfgInt("dpy_dungeon_height") / ref_pixels_per_square;

    static const int min_font_size = CfgInt("dpy_font_min_size");
    static const float ref_font_size = CfgFloat("dpy_font_base_size");

    // resize the dungeon
    int phy_dungeon_left, phy_dungeon_top, phy_pixels_per_square;
    Resize(ref_dungeon_left, ref_dungeon_top, ref_pixels_per_square,
           dungeon_tiles_x, dungeon_tiles_y,
           scale, left, top,
           phy_dungeon_left, phy_dungeon_top, phy_pixels_per_square);
    const int phy_dungeon_width = phy_pixels_per_square * dungeon_tiles_x;
    const int phy_dungeon_height = phy_pixels_per_square * dungeon_tiles_y;
    const float dungeon_scale_factor = float(phy_pixels_per_square) / float(ref_pixels_per_square);

    // set an appropriate font size, based on dungeon_scale_factor
    gm.setFontSize(max(min_font_size, Round(ref_font_size * dungeon_scale_factor)));
    
    // resize the inventory
    int phy_inventory_left, phy_inventory_top, phy_inventory_blk_sz;
    Resize(ref_inventory_left, ref_inventory_top, ref_inventory_width / num_inventory_slots,
           num_inventory_slots, 0,
           scale, left, top,
           phy_inventory_left, phy_inventory_top, phy_inventory_blk_sz);
    const int phy_inventory_width = phy_inventory_blk_sz * num_inventory_slots;
    const int phy_inventory_height = Round(float(ref_inventory_height) * scale);

    // resize the mini map
    // NOTE: assumption is that each square on the minimap takes up 1 pixel on the reference display.
    // If appropriate, we recentre that within the "ref_mini_map" rectangle.
    int phy_map_left = Round(float(ref_mini_map_left) * scale) + left;
    int phy_map_top = Round(float(ref_mini_map_top) * scale) + top;
    const int total_map_width = Round(float(ref_mini_map_width) * scale);
    const int total_map_height = Round(float(ref_mini_map_height) * scale);
    const int phy_map_width = Round(float(mini_map.getWidth()) * scale);
    const int phy_map_height = Round(float(mini_map.getHeight()) * scale);
    phy_map_left += (total_map_width - phy_map_width) / 2;
    phy_map_top += (total_map_height - phy_map_height) / 2;

    // Resize the menu. Assumption: menu gfx are square, and all of the same size.
    const int ref_menu_size = menu_gfx_empty ? menu_gfx_empty->getWidth() : 1;
    int phy_menu_left, phy_menu_top, phy_menu_size;
    Resize(ref_mini_map_left, ref_mini_map_top, ref_menu_size, 3, 3,
           scale, left, top,
           phy_menu_left, phy_menu_top, phy_menu_size);

    // simple calculation of screen extents (for screen flash)
    const int phy_scrn_left = min(phy_dungeon_left, phy_inventory_left);
    const int phy_scrn_right = max(phy_dungeon_left + phy_dungeon_width, phy_inventory_left + phy_inventory_width);
    const int phy_scrn_top = min(phy_dungeon_top, phy_inventory_top);
    const int phy_scrn_bottom = max(phy_dungeon_top + phy_dungeon_height, phy_inventory_top + phy_inventory_height);

    // If game is over, we do something different...
    if (won || lost) {
        if (won) {
            gc.drawText(phy_dungeon_left + 10, 50, *gm.getFont(), "WINNER", Coercri::Color(255,255,255), true);
        } else {
            gc.drawText(phy_dungeon_left + 10, 50, *gm.getFont(), "LOSER", Coercri::Color(255,255,255), true);
        }
        return;
    }

    // Clear out expired icons
    while (!icons.empty() && gvt > icons.top().expiry) {
        map<int,RoomData>::iterator ri = rooms.find(icons.top().room_no);
        if (ri != rooms.end()) {
            list<RoomData::GfxEntry> &glst(ri->second.lookupGfx(icons.top().x,
                                                                icons.top().y));
            for (list<RoomData::GfxEntry>::iterator it = glst.begin(); it != glst.end();
            ++it) {
                if (it->depth == icon_depth) {
                    glst.erase(it);
                    break;
                }
            }
        }
        icons.pop();
    }

    // Clear out expired messages
    // Note - this is not perfect as queue is ordered by start_time, rather than stop_time.
    // But this does not matter much in practice.
    while (!messages.empty() && gvt >= messages.front().stop_time) {
        messages.pop_front();
    }

    // Check whether screen should flash
    const bool screen_flash = (gvt >= flash_screen_start && gvt < flash_screen_start + CfgInt("screen_flash_time"));

    // Draw the coloured background if screen is flashing
    if (screen_flash)
    {
        Coercri::Color col(CfgInt("screen_flash_r"), CfgInt("screen_flash_g"), CfgInt("screen_flash_b"));
        Coercri::Rectangle rect(phy_scrn_left, phy_scrn_top, phy_scrn_right-phy_scrn_left, phy_scrn_bottom-phy_scrn_top);
        gc.fillRectangle(rect, col);
    }

    // Draw current room
    if (my_x != MY_X_OFFSCREEN_VALUE) last_known_time = gvt;
    bool alive_or_recently_deceased = gvt < last_known_time + CfgInt("death_draw_map_time");
    if (alive_or_recently_deceased) {
        // We are alive, or have been dead for no longer than "death_draw_map_time".
        // Draw the map.
        map<int,RoomData>::iterator ri = rooms.find(current_room);
        if (ri != rooms.end()) {
            RoomData &rd(ri->second);

            // Clear out the graphics buffer
            for (map<int, vector<GraphicElement> >::iterator it = gfx_buffer.begin();
            it != gfx_buffer.end(); ++it) {
                it->second.clear();
            }

            // Find on-screen top-left corner for the room
            const int room_tl_x = phy_dungeon_left + (phy_dungeon_width - rd.width*phy_pixels_per_square) / 2;
            const int room_tl_y = phy_dungeon_top + (phy_dungeon_height - rd.height*phy_pixels_per_square) / 2;

            // Build a list of graphics to be drawn

            // Draw tiles (if screen not flashing)
            if (!screen_flash) {
                for (int i=0; i<rd.width; ++i) {
                    for (int j=0; j<rd.height; ++j) {
                        const int screen_x = i*phy_pixels_per_square + room_tl_x;
                        const int screen_y = j*phy_pixels_per_square + room_tl_y;
                        
                        const list<RoomData::GfxEntry> &glst(rd.lookupGfx(i,j));
                        for (list<RoomData::GfxEntry>::const_iterator it = glst.begin();
                        it != glst.end(); ++it) {
                            GraphicElement ge;
                            ge.sx = screen_x;
                            ge.sy = screen_y;
                            ge.gr = it->gfx;
                            ge.cc = it->cc.get();
                            gfx_buffer[it->depth].push_back(ge);
                        }
                    }
                }
            }

            // Add the entities (always)
            rd.emap.getEntityGfx(gvt, room_tl_x, room_tl_y, phy_pixels_per_square, entity_depth, gfx_buffer);

            // Now draw all buffered graphics
            for (map<int, vector<GraphicElement> >::reverse_iterator it = gfx_buffer.rbegin();
            it != gfx_buffer.rend(); ++it) {
                for (vector<GraphicElement>::const_iterator it2 = it->second.begin();
                it2 != it->second.end(); ++it2) {

                    const int old_width = it2->gr->getWidth();
                    const int old_height = it2->gr->getHeight();
                    const int new_width = Round(old_width * dungeon_scale_factor);
                    const int new_height = Round(old_height * dungeon_scale_factor);

                    if (it2->cc) {
                        gm.drawTransformedGraphic(gc, it2->sx, it2->sy, *it2->gr, new_width, new_height, *it2->cc);
                    } else {
                        gm.drawTransformedGraphic(gc, it2->sx, it2->sy, *it2->gr, new_width, new_height);
                    }
                }
            }
        }
    } else {
        // Dead - clear out the msg queue as it looks stupid when msgs from your previous life come up!
        messages.clear();
        cts_messages.clear();
    }

    // The rest does not need to be drawn when screen is flashing
    if (!screen_flash) {

        using Coercri::Rectangle;
        
        // Draw inventory icons. (This is handled by DrawUI)
        for (map<int, BackpackEntry>::iterator it = backpack.begin(); it != backpack.end();
             ++it) {
            DrawUI::drawBackpackEntry(gc,
                                      phy_inventory_left, phy_inventory_top, phy_inventory_width, phy_inventory_height,
                                      gm, num_inventory_slots, ref_inventory_spacing, ref_inventory_gem_height,
                                      it->first, it->second.gfx, it->second.overdraw, it->second.no_carried,
                                      it->second.no_max);
        }
        
        // Draw Mini-Map or Menu
        if (menu_open) {
            // Menu is drawn in same position as mini-map.
            DrawUI::drawMenu(gc,
                             phy_menu_left, phy_menu_top, phy_menu_size*3, phy_menu_size*3,
                             gm, 
                             menu_gfx_centre, menu_gfx_empty, menu_gfx_highlight, gvt,
                             menu_control[D_NORTH]? menu_control[D_NORTH]->getMenuGraphic() : 0,
                             menu_control[D_EAST]? menu_control[D_EAST]->getMenuGraphic() : 0,
                             menu_control[D_SOUTH]? menu_control[D_SOUTH]->getMenuGraphic() : 0,
                             menu_control[D_WEST]? menu_control[D_WEST]->getMenuGraphic() : 0,
                             menu_highlight, menu_highlight_dir);
        } else {
            // Mini-Map is only drawn when menu is closed
            mini_map.draw(gc, phy_map_left, phy_map_top, phy_map_width, phy_map_height, gvt);
        }

        // Draw Skulls
        if (skull_renderer) {
            const int phy_skulls_left = Round(float(ref_skulls_left) * scale) + left;
            const int phy_skulls_top = Round(float(ref_skulls_top) * scale) + top;
            skull_renderer->draw(gc, gm, nskulls, phy_skulls_left, phy_skulls_top, scale);
        }

        // Draw Potion Bottle
        if (potion_renderer && alive_or_recently_deceased) {
            const int phy_potion_left = Round(float(ref_potion_left) * scale) + left;
            const int phy_potion_top = Round(float(ref_potion_top) * scale) + top;
            potion_renderer->draw(gvt, gc, gm, health, magic, poison_immun,
                                  phy_potion_left, phy_potion_top, scale);
        }

        // Determine whether we should draw a message
        if (messages.empty() && cts_messages.empty()) {
            last_mtime = 0;
        } else {
            if (last_mtime == 0) {
                if (!messages.empty()) last_mtime = messages.front().start_time;
                else last_mtime = gvt;
            }
            int mtime = gvt - last_mtime;

            static const int msg_on_time = CfgInt("message_on_time");
            static const int msg_off_time = CfgInt("message_off_time");
            const int nmsgs = messages.size() + cts_messages.size();
            int mphase = (mtime / (msg_on_time + msg_off_time)) % nmsgs;
            const string &the_msg(mphase < messages.size() ? messages[mphase].message
                                  : cts_messages[mphase - messages.size()]);
            if (mtime % (msg_on_time + msg_off_time) <= msg_on_time) {
                map<int,RoomData>::iterator ri = rooms.find(current_room);
                if (ri != rooms.end()) {
                    DrawUI::drawMessage(gc,
                                        phy_dungeon_left, phy_dungeon_top, phy_dungeon_width, phy_dungeon_height,
                                        gm,
                                        last_known_x, last_known_y, ri->second.width,
                                        ri->second.height, phy_pixels_per_square, the_msg);
                }
            }
        }
    }
}

